@@ -5,7 +5,7 @@
 
     Custom docutils writer for ReStructuredText.
 
-    :copyright: Copyright 2012-2013 by Freek Dijkstra.
+    :copyright: Copyright 2012-2021 by Freek Dijkstra and contributors.
     :license: BSD, see LICENSE.txt for details.
 """
 
@@ -13,14 +13,13 @@ from __future__ import (print_function, unicode_literals, absolute_import)
 
 import os
 import sys
-import re
 import textwrap
 import logging
 
 from docutils import nodes, writers
 
 from sphinx import addnodes
-from sphinx.locale import admonitionlabels, versionlabels, _
+from sphinx.locale import admonitionlabels, _
 from sphinx.writers.text import TextTranslator, MAXWIDTH, STDINDENT
 
 
@@ -58,13 +57,14 @@ class RstTranslator(TextTranslator):
         self.states = [[]]
         self.stateindent = [0]
         self.list_counter = []
+        self.list_formatter = []
         self.sectionlevel = 0
         self.table = None
         if self.builder.config.rst_indent:
             self.indent = self.builder.config.rst_indent
         else:
             self.indent = STDINDENT
-        self.wrapper = textwrap.TextWrapper(width=STDINDENT, break_long_words=False, break_on_hyphens=False)
+        self.wrapper = textwrap.TextWrapper(width=MAXWIDTH, break_long_words=False, break_on_hyphens=False)
 
     def log_unknown(self, type, node):
         logger = logging.getLogger("sphinxcontrib.writers.rst")
@@ -74,7 +74,7 @@ class RstTranslator(TextTranslator):
             logger = logging.getLogger("sphinxcontrib.writers.rst")
         logger.warning("%s(%s) unsupported formatting" % (type, node))
 
-    def wrap(self, text, width=STDINDENT):
+    def wrap(self, text, width=MAXWIDTH):
         self.wrapper.width = width
         return self.wrapper.wrap(text)
 
@@ -85,7 +85,7 @@ class RstTranslator(TextTranslator):
         self.stateindent.append(indent)
     def end_state(self, wrap=True, end=[''], first=None):
         content = self.states.pop()
-        maxindent = sum(self.stateindent)
+        width = max(MAXWIDTH//3, MAXWIDTH - sum(self.stateindent))
         indent = self.stateindent.pop()
         result = []
         toformat = []
@@ -93,7 +93,7 @@ class RstTranslator(TextTranslator):
             if not toformat:
                 return
             if wrap:
-                res = self.wrap(''.join(toformat), width=MAXWIDTH-maxindent)
+                res = self.wrap(''.join(toformat), width=width)
             else:
                 res = ''.join(toformat).splitlines()
             if end:
@@ -458,9 +458,16 @@ class RstTranslator(TextTranslator):
         raise nodes.SkipNode
 
     def visit_image(self, node):
-        if 'alt' in node.attributes:
+        self.new_state(0)
+        if 'uri' in node:
+            self.add_text(_('.. image:: %s') % node['uri'])
+        elif 'target' in node.attributes:
+            self.add_text(_('.. image: %s') % node['target'])
+        elif 'alt' in node.attributes:
             self.add_text(_('[image: %s]') % node['alt'])
-        self.add_text(_('[image]'))
+        else:
+            self.add_text(_('[image]'))
+        self.end_state(wrap=False)
         raise nodes.SkipNode
 
     def visit_transition(self, node):
@@ -471,38 +478,40 @@ class RstTranslator(TextTranslator):
         raise nodes.SkipNode
 
     def visit_bullet_list(self, node):
-        self.list_counter.append(-1)
+        def bullet_list_format(counter):
+            return '*'
+        self.list_counter.append(-1)  # TODO: just 0 is fine.
+        self.list_formatter.append(bullet_list_format)
     def depart_bullet_list(self, node):
         self.list_counter.pop()
+        self.list_formatter.pop()
 
     def visit_enumerated_list(self, node):
+        def enumerated_list_format(counter):
+            return str(counter) + '.'
         self.list_counter.append(0)
+        self.list_formatter.append(enumerated_list_format)
     def depart_enumerated_list(self, node):
         self.list_counter.pop()
-
-    def visit_definition_list(self, node):
-        self.list_counter.append(-2)
-    def depart_definition_list(self, node):
-        self.list_counter.pop()
+        self.list_formatter.pop()
 
     def visit_list_item(self, node):
-        if self.list_counter[-1] == -1:
-            # bullet list
-            self.new_state(self.indent)
-        elif self.list_counter[-1] == -2:
-            # definition list
-            pass
-        else:
-            # enumerated list
-            self.list_counter[-1] += 1
-            self.new_state(len(str(self.list_counter[-1])) + self.indent)
+        self.list_counter[-1] += 1
+        bullet_formatter = self.list_formatter[-1]
+        bullet = bullet_formatter(self.list_counter[-1])
+        indent = max(self.indent, len(bullet) + 1)
+        self.new_state(indent)
     def depart_list_item(self, node):
-        if self.list_counter[-1] == -1:
-            self.end_state(first='* ', end=None)
-        elif self.list_counter[-1] == -2:
-            pass
-        else:
-            self.end_state(first='%s. ' % self.list_counter[-1], end=None)
+        # formatting to make the string `self.stateindent[-1]` chars long.
+        format = '%%-%ds' % (self.stateindent[-1])
+        bullet_formatter = self.list_formatter[-1]
+        bullet = format % bullet_formatter(self.list_counter[-1])
+        self.end_state(first=bullet, end=None)
+
+    def visit_definition_list(self, node):
+        pass
+    def depart_definition_list(self, node):
+        pass
 
     def visit_definition_list_item(self, node):
         self._li_has_classifier = len(node) >= 2 and \
@@ -603,16 +612,34 @@ class RstTranslator(TextTranslator):
 
     def visit_versionmodified(self, node):
         self.new_state(0)
-        if node.children:
-            self.add_text(versionlabels[node['type']] % node['version'] + ': ')
-        else:
-            self.add_text(versionlabels[node['type']] % node['version'] + '.')
     def depart_versionmodified(self, node):
         self.end_state()
 
     def visit_literal_block(self, node):
-        self.add_text("::")
+        is_code_block = False
+        # Support for Sphinx < 2.0, which defines classes['code', 'language']
+        if 'code' in node.get('classes', []):
+            is_code_block = True
+            if node.get('language', 'default') == 'default' and len(node['classes']) >= 2:
+                node['language'] = node['classes'][1]
+        # highlight_args is the only way to distinguish between :: and .. code:: in Sphinx 2 or higher.
+        if node.get('highlight_args') != None:
+            is_code_block = True
+        if is_code_block:
+            if node.get('language', 'default') == 'default':
+                directive = ".. code::"
+            else:
+                directive = ".. code:: %s" % node['language']
+            if node.get('linenos'):
+                indent = self.indent * ' '
+                directive += "%s%s:number-lines:" % (self.nl, indent)
+        else:
+            directive = "::"
+        self.new_state(0)
+        self.add_text(directive)
+        self.end_state(wrap=False)
         self.new_state(self.indent)
+
     def depart_literal_block(self, node):
         self.end_state(wrap=False)
 
@@ -633,7 +660,6 @@ class RstTranslator(TextTranslator):
         pass
 
     def visit_block_quote(self, node):
-        self.add_text('..')
         self.new_state(self.indent)
     def depart_block_quote(self, node):
         self.end_state()
@@ -655,7 +681,12 @@ class RstTranslator(TextTranslator):
     def visit_target(self, node):
         if 'refid' in node:
             self.new_state(0)
-            self.add_text('.. _'+node['refid']+':'+self.nl)
+            if node.get('ids'):
+                self.add_text(self.nl.join(
+                    '.. _%s: %s_' % (id, node['refid']) for id in node['ids']
+                ))
+            else:
+                self.add_text('.. _'+node['refid']+':')
     def depart_target(self, node):
         if 'refid' in node:
             self.end_state(wrap=False)
@@ -672,63 +703,38 @@ class RstTranslator(TextTranslator):
         pass
 
     def visit_reference(self, node):
-        """Run upon entering a reference
-
-        Because this class inherits from the TextTranslator class,
-        regularly defined links, such as::
-
-            `Some Text`_
-
-            .. _Some Text: http://www.some_url.com
-
-        were being written as plaintext. This included internal
-        references defined in the standard rst way, such as::
-
-            `Some Reference`
-
-            .. _Some Reference:
-
-            Some Title
-            ----------
-
-        To resolve this, if ``refuri`` is not included in the node (an
-        internal, non-Sphinx-defined internal uri, the reference is
-        left unchanged (e.g. ```Some Text`_`` is written as such).
-
-        If ``internal`` is not in the node (as for an external,
-        non-Sphinx URI, the reference is rewritten as an inline link,
-        e.g. ```Some Text <http://www.some_url.com>`_``.
-
-        If ``reftitle` is in the node (as in a Sphinx-generated
-        reference), the node is converted to an inline link.
-
-        Finally, all other links are also converted to an inline link
-        format.
-        """
-        if 'refuri' not in node:
-            self.add_text('`%s`_' % node['name'])
-            raise nodes.SkipNode
-        elif 'internal' not in node:
-            self.add_text('`%s <%s>`_' % (node['name'], node['refuri']))
-            raise nodes.SkipNode
-        elif 'reftitle' in node:
-            # Include node as text, rather than with markup.
-            # reST seems unable to parse a construct like ` ``literal`` <url>`_
-            # Hence we revert to the more simple `literal <url>`_
-            self.add_text('`%s <%s>`_' % (node.astext(), node['refuri']))
-            # self.end_state(wrap=False)
-            raise nodes.SkipNode
+        refname = node.get('name')
+        refbody = node.astext()
+        refuri = node.get('refuri')
+        refid = node.get('refid')
+        if node.get('anonymous'):
+            underscore = '__'
         else:
-            self.add_text('`%s <%s>`_' % (node.astext(), node['refuri']))
-            raise nodes.SkipNode
+            underscore = '_'
+
+        if refuri and refuri.endswith('_'):
+            refuri = refuri[:-1] + '\\_'
+
+        if node.get('internal'):
+            if refuri:
+                self.add_text('`%s <%s>`%s' % (refbody, refuri, underscore))
+            elif refid:
+                self.add_text('`%s <#%s>`%s' % (refbody, refid, underscore))
+            else:
+                return
+        else:
+            if refuri and refname:
+                self.add_text('`%s <%s>`%s' % (refname, refuri, underscore))
+            elif refuri:
+                self.add_text('%s' % (refuri))
+            elif refname:
+                self.add_text('`%s`%s' % (refname, underscore))
+            else:
+                return
+        raise nodes.SkipNode
 
     def depart_reference(self, node):
-        if 'refuri' not in node:
-            pass # Don't add these anchors
-        elif 'internal' not in node:
-            pass # Don't add external links (they are automatically added by the reST spec)
-        elif 'reftitle' in node:
-            pass
+        pass
 
     def visit_download_reference(self, node):
         self.log_unknown("download_reference", node)
@@ -769,14 +775,16 @@ class RstTranslator(TextTranslator):
         self.add_text('``')
 
     def visit_subscript(self, node):
-        self.add_text('_')
+        self.add_text(':sub:`')
+
     def depart_subscript(self, node):
-        pass
+        self.add_text('`')
 
     def visit_superscript(self, node):
-        self.add_text('^')
+        self.add_text(':sup:`')
+
     def depart_superscript(self, node):
-        pass
+        self.add_text('`')
 
     def visit_footnote_reference(self, node):
         self.add_text('[%s]' % node.astext())
@@ -804,9 +812,9 @@ class RstTranslator(TextTranslator):
         pass
 
     def visit_problematic(self, node):
-        self.add_text('>>')
+        pass
     def depart_problematic(self, node):
-        self.add_text('<<')
+        pass
 
     def visit_system_message(self, node):
         self.new_state(0)
